<?php
namespace nur\sery\php\time;

use DateTimeInterface;
use DateTimeZone;
use InvalidArgumentException;

/**
 * Class DateTime: une date et une heure
 *
 * @property-read int $year
 * @property-read int $month
 * @property-read int $day
 * @property-read int $hour
 * @property-read int $minute
 * @property-read int $second
 * @property-read int $wday
 * @property-read int $wnum
 * @property-read string $timezone
 * @property-read string $datetime
 * @property-read string $date
 */
class DateTime extends \DateTime {
  const DEFAULT_FORMAT = "d/m/Y H:i:s";
  const INT_FORMATS = [
    "year" => "Y",
    "month" => "m",
    "day" => "d",
    "hour" => "H",
    "minute" => "i",
    "second" => "s",
    "wday" => "N",
    "wnum" => "W",
  ];
  const STRING_FORMATS = [
    "timezone" => "P",
    "datetime" => "d/m/Y H:i:s",
    "date" => "d/m/Y",
  ];

  static function clone(DateTimeInterface $dateTime): self {
    if ($dateTime instanceof static) return clone $dateTime;
    $clone = new static();
    $clone->setTimestamp($dateTime->getTimestamp());
    $clone->setTimezone($dateTime->getTimezone());
    return $clone;
  }

  /**
   * corriger une année à deux chiffres qui est située dans le passé et
   * retourner l'année à 4 chiffres.
   *
   * par exemple, si l'année courante est 2019, alors:
   * - fix_past_year('18') === '2018'
   * - fix_past_year('19') === '1919'
   * - fix_past_year('20') === '1920'
   */
  static function fix_past_year(int $year): int {
    if ($year < 100) {
      $y = getdate(); $y = $y["year"];
      $r = $y % 100;
      $c = $y - $r;
      if ($year >= $r) $year += $c - 100;
      else $year += $c;
    }
    return $year;
  }

  /**
   * corriger une année à deux chiffres et retourner l'année à 4 chiffres.
   * l'année charnière entre année passée et année future est 70
   *
   * par exemple, si l'année courante est 2019, alors:
   * - fix_past_year('18') === '2018'
   * - fix_past_year('19') === '2019'
   * - fix_past_year('20') === '2020'
   * - fix_past_year('69') === '2069'
   * - fix_past_year('70') === '1970'
   * - fix_past_year('71') === '1971'
   */
  static function fix_any_year(int $year): int {
    if ($year < 100) {
      $y = intval(date("Y"));
      $r = $y % 100;
      $c = $y - $r;
      if ($year >= 70) $year += $c - 100;
      else $year += $c;
    }
    return $year;
  }

  function __construct($datetime="now", DateTimeZone $timezone=null) {
    if ($datetime instanceof \DateTimeInterface) {
      if ($timezone === null) $timezone = $datetime->getTimezone();
      parent::__construct();
      $this->setTimestamp($datetime->getTimestamp());
      $this->setTimezone($timezone);
    } elseif (is_int($datetime)) {
      parent::__construct("now", $timezone);
      $this->setTimestamp($datetime);
    } elseif (!is_string($datetime)) {
      throw new InvalidArgumentException("datetime must be a string or an instance of DateTimeInterface");
    } else {
      if (preg_match('/^(\d+)\/(\d+)\/(\d+)$/', $datetime, $ms)) {
        $y = self::fix_any_year(intval($ms[3]));
        $datetime = sprintf("%04d-%02d-%02d", $y, intval($ms[2]), intval($ms[1]));
      } elseif (preg_match('/^(\d+)\/(\d+)$/', $datetime, $ms)) {
        $datetime = sprintf("%04d-%02d-%02d", intval(date("Y")), intval($ms[2]), intval($ms[1]));
      } elseif (preg_match('/^(\d{4})(\d{2})(\d{2})$/', $datetime, $ms)) {
        $datetime = sprintf("%04d-%02d-%02d", intval($ms[1]), intval($ms[2]), intval($ms[3]));
      } elseif (preg_match('/^(\d{2})(\d{2})(\d{2})$/', $datetime, $ms)) {
        $y = self::fix_any_year(intval($ms[1]));
        $datetime = sprintf("%04d-%02d-%02d", $y, intval($ms[2]), intval($ms[3]));
      }
      #XXX ajouter datetime et YYYYmmddTHHMMSS[Z]
      parent::__construct($datetime, $timezone);
    }
  }

  function diff($target, $absolute=false): DateInterval {
    return new DateInterval(parent::diff($target, $absolute));
  }

  function format($format=self::DEFAULT_FORMAT): string {
    return \DateTime::format($format);
  }

  function __toString(): string {
    return $this->format();
  }

  function __get($name) {
    if (array_key_exists($name, self::INT_FORMATS)) {
      return intval($this->format(self::INT_FORMATS[$name]));
    } elseif (array_key_exists($name, self::STRING_FORMATS)) {
      return $this->format(self::STRING_FORMATS[$name]);
    }
    throw new InvalidArgumentException("Unknown property $name");
  }
}
